@@ -55,6 +55,8 @@ struct _GskContourClass
                                                  GString                *string);
   gboolean              (* get_bounds)          (const GskContour       *contour,
                                                  GskBoundingBox         *bounds);
+  gboolean              (* get_tight_bounds)    (const GskContour       *contour,
+                                                 GskBoundingBox         *bounds);
   gboolean              (* get_stroke_bounds)   (const GskContour       *contour,
                                                  const GskStroke        *stroke,
                                                  GskBoundingBox         *bounds);
@@ -678,6 +680,43 @@ gsk_standard_contour_get_bounds (const GskContour *contour,
   return bounds->max.x > bounds->min.x && bounds->max.y > bounds->min.y;
 }
 
+static gboolean
+add_tight_bounds (GskPathOperation        op,
+                  const graphene_point_t *pts,
+                  gsize                   n_pts,
+                  float                   weight,
+                  gpointer                user_data)
+{
+  GskBoundingBox *bounds = user_data;
+  GskCurve c;
+  GskBoundingBox b;
+
+  if (op == GSK_PATH_MOVE)
+    return TRUE;
+
+  gsk_curve_init_foreach (&c, op, pts, n_pts, weight);
+  gsk_curve_get_tight_bounds (&c, &b);
+  gsk_bounding_box_union (&b, bounds, bounds);
+
+  return TRUE;
+}
+
+static gboolean
+gsk_standard_contour_get_tight_bounds (const GskContour *contour,
+                                       GskBoundingBox   *bounds)
+{
+  const GskStandardContour *self = (const GskStandardContour *) contour;
+
+  if (self->n_points == 0)
+    return FALSE;
+
+  gsk_bounding_box_init (bounds, &self->points[0].pt, &self->points[0].pt);
+  for (gsize i = 0; i < self->n_ops; i ++)
+    gsk_pathop_foreach (self->ops[i], add_tight_bounds, bounds);
+
+  return bounds->max.x > bounds->min.x && bounds->max.y > bounds->min.y;
+}
+
 static gboolean
 gsk_standard_contour_get_stroke_bounds (const GskContour *contour,
                                         const GskStroke  *stroke,
@@ -990,13 +1029,17 @@ add_measure (const GskCurve *curve,
              float           tolerance,
              float           t1,
              float           l1,
-             GArray         *array)
+             GArray         *array,
+             int             depth)
 {
   GskCurve c;
   float ll, l0;
   float t0;
   CurvePoint *p;
 
+  if (depth > 10)
+    goto done;
+
   /* Check if we can add (t1, length + l1) without further
    * splitting. We check two things:
    * - Is the curve close to a straight line (length-wise) ?
@@ -1024,8 +1067,8 @@ done:
     }
   else
     {
-      add_measure (curve, idx, length, tolerance, t0, l0, array);
-      add_measure (curve, idx, length, tolerance, t1, l1, array);
+      add_measure (curve, idx, length, tolerance, t0, l0, array, depth + 1);
+      add_measure (curve, idx, length, tolerance, t1, l1, array, depth + 1);
     }
 }
 
@@ -1066,10 +1109,10 @@ add_samples (const GskStandardContour  *self,
   for (int j = 0; j < n; j++)
     {
       float l = gsk_curve_get_length_to (&curve, t[j]);
-      add_measure (&curve, curve_measure->idx, l0, measure->tolerance, t[j], l, measure->points);
+      add_measure (&curve, curve_measure->idx, l0, measure->tolerance, t[j], l, measure->points, 0);
     }
 
-  add_measure (&curve, curve_measure->idx, l0, measure->tolerance, 1, l1 - l0, measure->points);
+  add_measure (&curve, curve_measure->idx, l0, measure->tolerance, 1, l1 - l0, measure->points, 0);
 
   curve_measure->first = first;
   curve_measure->n_samples = measure->points->len - first;
@@ -1303,6 +1346,7 @@ static const GskContourClass GSK_STANDARD_CONTOUR_CLASS =
   gsk_standard_contour_get_flags,
   gsk_contour_print_default,
   gsk_standard_contour_get_bounds,
+  gsk_standard_contour_get_tight_bounds,
   gsk_standard_contour_get_stroke_bounds,
   gsk_standard_contour_foreach,
   gsk_standard_contour_reverse,
@@ -1377,6 +1421,38 @@ gsk_standard_contour_new (GskPathFlags            flags,
   return contour;
 }
 
+size_t
+gsk_contour_get_standard_ops (const GskContour *contour,
+                              size_t            n_ops,
+                              GskPathOperation *ops)
+{
+  const GskStandardContour *self = (const GskStandardContour *) contour;
+
+  if (contour->klass != &GSK_STANDARD_CONTOUR_CLASS)
+    return 0;
+
+  for (size_t i = 0; i < MIN (n_ops, self->n_ops); i++)
+    ops[i] = gsk_pathop_op (self->ops[i]);
+
+  return self->n_ops;
+}
+
+size_t
+gsk_contour_get_standard_points (const GskContour *contour,
+                                 size_t            n_points,
+                                 graphene_point_t *points)
+{
+  const GskStandardContour *self = (const GskStandardContour *) contour;
+
+  if (contour->klass != &GSK_STANDARD_CONTOUR_CLASS)
+    return 0;
+
+  for (size_t i = 0; i < MIN (n_points, self->n_points); i++)
+    points[i] = self->points[i].pt;
+
+  return self->n_points;
+}
+
 /* }}} */
 /* {{{ Circle */
 
@@ -1403,7 +1479,10 @@ gsk_circle_contour_copy (const GskContour *contour,
 static GskPathFlags
 gsk_circle_contour_get_flags (const GskContour *contour)
 {
-  return GSK_PATH_CLOSED;
+  const GskCircleContour *self = (const GskCircleContour *) contour;
+
+  return GSK_PATH_CLOSED |
+         (self->radius == 0 ? GSK_PATH_ZERO_LENGTH : 0);
 }
 
 static void
@@ -1783,6 +1862,7 @@ static const GskContourClass GSK_CIRCLE_CONTOUR_CLASS =
   gsk_circle_contour_get_flags,
   gsk_circle_contour_print,
   gsk_circle_contour_get_bounds,
+  gsk_circle_contour_get_bounds,
   gsk_circle_contour_get_stroke_bounds,
   gsk_circle_contour_foreach,
   gsk_circle_contour_reverse,
@@ -1812,7 +1892,6 @@ gsk_circle_contour_new (const graphene_point_t *center,
 
   self->contour.klass = &GSK_CIRCLE_CONTOUR_CLASS;
 
-  self->contour.klass = &GSK_CIRCLE_CONTOUR_CLASS;
   self->center = *center;
   self->radius = radius;
   self->ccw = FALSE;
@@ -1820,17 +1899,22 @@ gsk_circle_contour_new (const graphene_point_t *center,
   return (GskContour *) self;
 }
 
-void
-gsk_circle_contour_get_params (const GskContour *contour,
-                               graphene_point_t *center,
-                               float            *radius,
-                               gboolean         *ccw)
+gboolean
+gsk_contour_get_circle (const GskContour *contour,
+                        graphene_point_t *center,
+                        float            *radius,
+                        gboolean         *ccw)
 {
   const GskCircleContour *self = (const GskCircleContour *) contour;
 
+  if (contour->klass != &GSK_CIRCLE_CONTOUR_CLASS)
+    return FALSE;
+
   *center = self->center;
   *radius = self->radius;
   *ccw = self->ccw;
+
+  return TRUE;
 }
 
 /* }}} */
@@ -1862,7 +1946,11 @@ gsk_rect_contour_copy (const GskContour *contour,
 static GskPathFlags
 gsk_rect_contour_get_flags (const GskContour *contour)
 {
-  return GSK_PATH_FLAT | GSK_PATH_CLOSED;
+  const GskRectContour *self = (const GskRectContour *) contour;
+
+  return GSK_PATH_FLAT |
+         GSK_PATH_CLOSED |
+         (self->width == 0 && self->height == 0 ? GSK_PATH_ZERO_LENGTH : 0);
 }
 
 static void
@@ -2143,6 +2231,7 @@ static const GskContourClass GSK_RECT_CONTOUR_CLASS =
   gsk_rect_contour_get_flags,
   gsk_rect_contour_print,
   gsk_rect_contour_get_bounds,
+  gsk_rect_contour_get_bounds,
   gsk_rect_contour_get_stroke_bounds,
   gsk_rect_contour_foreach,
   gsk_rect_contour_reverse,
@@ -2179,6 +2268,23 @@ gsk_rect_contour_new (const graphene_rect_t *rect)
   return (GskContour *) self;
 }
 
+gboolean
+gsk_contour_get_rect (const GskContour *contour,
+                      graphene_rect_t  *rect)
+{
+  const GskRectContour *self = (const GskRectContour *) contour;
+
+  if (contour->klass != &GSK_RECT_CONTOUR_CLASS)
+    return FALSE;
+
+  rect->origin.x = self->x;
+  rect->origin.y = self->y;
+  rect->size.width = self->width;
+  rect->size.height = self->height;
+
+  return TRUE;
+}
+
 /* }}} */
 /* {{{ Rounded Rectangle */
 
@@ -2219,6 +2325,21 @@ gsk_rounded_rect_contour_get_bounds (const GskContour *contour,
   return TRUE;
 }
 
+static gboolean
+gsk_rounded_rect_contour_get_tight_bounds (const GskContour *contour,
+                                           GskBoundingBox   *bounds)
+{
+  GskPath *path;
+  graphene_rect_t b;
+  gboolean ret;
+
+  path = convert_to_standard_contour (contour);
+  ret = gsk_path_get_tight_bounds (path, &b);
+  gsk_bounding_box_init_from_rect (bounds, &b);
+  gsk_path_unref (path);
+
+  return ret;
+}
 static gboolean
 gsk_rounded_rect_contour_get_stroke_bounds (const GskContour *contour,
                                             const GskStroke  *stroke,
@@ -2489,6 +2610,7 @@ static const GskContourClass GSK_ROUNDED_RECT_CONTOUR_CLASS =
   gsk_rounded_rect_contour_get_flags,
   gsk_contour_print_default,
   gsk_rounded_rect_contour_get_bounds,
+  gsk_rounded_rect_contour_get_tight_bounds,
   gsk_rounded_rect_contour_get_stroke_bounds,
   gsk_rounded_rect_contour_foreach,
   gsk_rounded_rect_contour_reverse,
@@ -2564,6 +2686,20 @@ gsk_rounded_rect_contour_new (const GskRoundedRect *rect)
   return (GskContour *) self;
 }
 
+gboolean
+gsk_contour_get_rounded_rect (const GskContour *contour,
+                              GskRoundedRect   *rect)
+{
+  const GskRoundedRectContour *self = (const GskRoundedRectContour *) contour;
+
+  if (contour->klass != &GSK_ROUNDED_RECT_CONTOUR_CLASS)
+    return FALSE;
+
+  *rect = self->rect;
+
+  return TRUE;
+}
+
 /* }}} */
 /* {{{ API */
 
@@ -2623,6 +2759,13 @@ gsk_contour_get_bounds (const GskContour *self,
   return self->klass->get_bounds (self, bounds);
 }
 
+gboolean
+gsk_contour_get_tight_bounds (const GskContour *self,
+                              GskBoundingBox   *bounds)
+{
+  return self->klass->get_tight_bounds (self, bounds);
+}
+
 gboolean
 gsk_contour_get_stroke_bounds (const GskContour *self,
                                const GskStroke  *stroke,
